const Meta = imports.gi.Meta;
const GLib = imports.gi.GLib;

const change_workspace = (win, manager, index) => {
  const n = manager.get_n_workspaces();
  if (n <= index) {
    return;
  }
  win.change_workspace_by_index(index, 1);
  manager.get_workspace_by_index(index).activate(global.get_current_time());
};

const _old_workspaces = {};

const first_empty_workspace_index = (manager, win) => {
  const n = manager.get_n_workspaces();
  let lastworkspace = n - 1;
  for (let i = 0; i < lastworkspace; ++i) {
    let win_count = manager.get_workspace_by_index(i)
      .list_windows()
      .filter(w => !w.is_always_on_all_workspaces() && win.get_monitor() == w.get_monitor()).length;
    if (win_count < 1) {
      return i;
    }
  }
  if (lastworkspace < 1) lastworkspace = 1
  return lastworkspace;
}

function check(win, change) {
  const workspacemanager = win.get_display().get_workspace_manager();
  if (win.window_type !== Meta.WindowType.NORMAL) {
    return;
  }
  let w = win.get_workspace().list_windows()
    .filter(w => w !== win && !w.is_always_on_all_workspaces() && win.get_monitor() == w.get_monitor());
  if (change === Meta.SizeChange.UNFULLSCREEN || change === Meta.SizeChange.UNMAXIMIZE || (change === Meta.SizeChange.MAXIMIZE && win.get_maximized() !== Meta.MaximizeFlags.BOTH)) {
    let name = win.get_id();
    if (_old_workspaces[name] !== undefined) {
      if (w.length == 0) { // TODO expose this as user choice
        change_workspace(win, workspacemanager, _old_workspaces[name]);
      }
      _old_workspaces[name] = undefined;
    }
    return;
  }
  if (w.length >= 1) {
    let emptyworkspace = first_empty_workspace_index(workspacemanager, win);

    if (emptyworkspace == win.get_workspace().index())
      return;

    let name = win.get_id();
    _old_workspaces[name] = win.get_workspace().index();
    change_workspace(win, workspacemanager, emptyworkspace);
  }
}

function checkFullScreen(win) {
  if (!win || !win.get_workspace()) {
    return false;
  }
  var windowActor = win.get_compositor_private();
  if (!windowActor || !windowActor.visible || !windowActor.has_allocation()) {
    return false;
  }
  if (win.get_maximized() === Meta.MaximizeFlags.BOTH &&
    _old_workspaces[win.get_id()] === undefined &&
    win.has_focus()) {
    check(win, Meta.SizeChange.MAXIMIZE);
  }
  return false;
}

function handleWindowClose(act) {
    let win = act.meta_window;
    let name = win.get_id();
    if (_old_workspaces[name] !== undefined) {
      win.get_display().get_workspace_manager().get_workspace_by_index(_old_workspaces[name]).activate(global.get_current_time());
    }
};

const _window_manager_handles = [];
const _display_handles = [];

function enable() {
  _display_handles.push(global.display.connect('window-created', (_, win) => {
    if (win.get_layer() !== Meta.StackLayer.NORMAL) {
      return;
    }
    if (win.get_frame_type() !== Meta.FrameType.NORMAL && win.get_frame_type() !== Meta.FrameType.BORDER) {
      return;
    }
    GLib.timeout_add(GLib.PRIORITY_LOW, 300, checkFullScreen.bind(this, win));
  }));
  _window_manager_handles.push(global.window_manager.connect('size-change', (_, act, change) => {
    check(act.meta_window, change);
  }));
  _window_manager_handles.push(global.window_manager.connect('destroy', (_, act) => {
    handleWindowClose(act);
  }));
}

function disable() {
  _window_manager_handles.splice(0).forEach(h => global.window_manager.disconnect(h));
  _display_handles.splice(0).forEach(h => global.display.disconnect(h));
}